@@ -24,115 +24,126 @@
 #include "dat.h"
 #include "hash.h"
 #include "string_in.h"
+#include "plugin_in.h"
 #include <groonga/tokenizer.h>
 
-grn_obj *grn_uvector_tokenizer = NULL;
+grn_obj *grn_token_uvector = NULL;
 
 typedef struct {
+  grn_tokenizer_token token;
   byte *curr;
   byte *tail;
   uint32_t unit;
-  grn_obj curr_;
-  grn_obj stat_;
-} grn_uvector_tokenizer_info;
+} grn_uvector_tokenizer;
 
 static grn_obj *
 uvector_init(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data)
 {
-  grn_obj *str;
-  grn_uvector_tokenizer_info *token;
-  if (!(str = grn_ctx_pop(ctx))) {
-    ERR(GRN_INVALID_ARGUMENT, "missing argument");
+  grn_obj *str, *flags;
+  grn_uvector_tokenizer *tokenizer;
+  if (!(flags = grn_ctx_pop(ctx))) {
+    ERR(GRN_INVALID_ARGUMENT, "[tokenizer][uvector] missing argument: flags");
     return NULL;
   }
-  if (!(token = GRN_MALLOC(sizeof(grn_uvector_tokenizer_info)))) { return NULL; }
-  user_data->ptr = token;
-  token->curr = GRN_TEXT_VALUE(str);
-  token->tail = token->curr + GRN_TEXT_LEN(str);
-  token->unit = sizeof(grn_id);
-  GRN_TEXT_INIT(&token->curr_, GRN_OBJ_DO_SHALLOW_COPY);
-  GRN_UINT32_INIT(&token->stat_, 0);
+  if (!(str = grn_ctx_pop(ctx))) {
+    ERR(GRN_INVALID_ARGUMENT, "[tokenizer][uvector] missing argument: string");
+    return NULL;
+  }
+  if (!(tokenizer = GRN_MALLOC(sizeof(grn_uvector_tokenizer)))) {
+    ERR(GRN_NO_MEMORY_AVAILABLE,
+        "[tokenizer][uvector] "
+        "memory allocation to grn_uvector_tokenizer failed");
+    return NULL;
+  }
+  user_data->ptr = tokenizer;
+
+  grn_tokenizer_token_init(ctx, &(tokenizer->token));
+  tokenizer->curr = (byte *)GRN_TEXT_VALUE(str);
+  tokenizer->tail = tokenizer->curr + GRN_TEXT_LEN(str);
+  tokenizer->unit = sizeof(grn_id);
   return NULL;
 }
 
 static grn_obj *
 uvector_next(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data)
 {
-  grn_uvector_tokenizer_info *token = user_data->ptr;
-  byte *p = token->curr + token->unit;
-  if (token->tail < p) {
-    GRN_TEXT_SET_REF(&token->curr_, token->curr, 0);
-    GRN_UINT32_SET(ctx, &token->stat_, GRN_TOKEN_LAST);
+  grn_uvector_tokenizer *tokenizer = user_data->ptr;
+  byte *p = tokenizer->curr + tokenizer->unit;
+  if (tokenizer->tail < p) {
+    grn_tokenizer_token_push(ctx, &(tokenizer->token),
+                             (const char *)tokenizer->curr, 0,
+                             GRN_TOKENIZER_TOKEN_LAST);
   } else {
-    GRN_TEXT_SET_REF(&token->curr_, token->curr, token->unit);
-    token->curr = p;
-    GRN_UINT32_SET(ctx, &token->stat_, token->tail == p ? GRN_TOKEN_LAST : 0);
+    grn_tokenizer_status status;
+    if (tokenizer->tail == p) {
+      status = GRN_TOKENIZER_TOKEN_LAST;
+    } else {
+      status = GRN_TOKENIZER_TOKEN_CONTINUE;
+    }
+    grn_tokenizer_token_push(ctx, &(tokenizer->token),
+                             (const char *)tokenizer->curr, tokenizer->unit,
+                             status);
+    tokenizer->curr = p;
   }
-  grn_ctx_push(ctx, &token->curr_);
-  grn_ctx_push(ctx, &token->stat_);
   return NULL;
 }
 
 static grn_obj *
 uvector_fin(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data)
 {
-  GRN_FREE(user_data->ptr);
+  grn_uvector_tokenizer *tokenizer = user_data->ptr;
+  if (!tokenizer) {
+    return NULL;
+  }
+  grn_tokenizer_token_fin(ctx, &(tokenizer->token));
+  GRN_FREE(tokenizer);
   return NULL;
 }
 
 typedef struct {
-  grn_obj *nstr;
   const uint8_t *delimiter;
   uint32_t delimiter_len;
-  grn_encoding encoding;
   const unsigned char *next;
   const unsigned char *end;
   grn_tokenizer_token token;
+  grn_tokenizer_query *query;
   grn_bool have_tokenized_delimiter;
 } grn_delimited_tokenizer;
 
 static grn_obj *
-delimited_init(grn_ctx *ctx, grn_obj *table, grn_user_data *user_data,
+delimited_init(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data,
                const uint8_t *delimiter, uint32_t delimiter_len)
 {
-  grn_obj *str;
-  grn_obj *normalizer = NULL;
-  int nflags = 0;
+  grn_tokenizer_query *query;
+  unsigned int normalize_flags = 0;
   const char *normalized;
   unsigned int normalized_length_in_bytes;
   grn_delimited_tokenizer *tokenizer;
-  grn_obj_flags table_flags;
-  if (!(str = grn_ctx_pop(ctx))) {
-    ERR(GRN_INVALID_ARGUMENT, "missing argument");
+
+  query = grn_tokenizer_query_open(ctx, nargs, args, normalize_flags);
+  if (!query) {
     return NULL;
   }
+
   if (!(tokenizer = GRN_MALLOC(sizeof(grn_delimited_tokenizer)))) {
+    ERR(GRN_NO_MEMORY_AVAILABLE,
+        "[tokenizer][delimit] "
+        "memory allocation to grn_delimited_tokenizer failed");
+    grn_tokenizer_query_close(ctx, query);
     return NULL;
   }
   user_data->ptr = tokenizer;
 
-  grn_table_get_info(ctx, table, &table_flags, &tokenizer->encoding, NULL);
+  tokenizer->query = query;
 
   tokenizer->have_tokenized_delimiter =
     grn_tokenizer_have_tokenized_delimiter(ctx,
-                                           GRN_TEXT_VALUE(str),
-                                           GRN_TEXT_LEN(str),
-                                           tokenizer->encoding);
+                                           tokenizer->query->ptr,
+                                           tokenizer->query->length,
+                                           tokenizer->query->encoding);
   tokenizer->delimiter = delimiter;
   tokenizer->delimiter_len = delimiter_len;
-
-  if (table_flags & GRN_OBJ_KEY_NORMALIZE) {
-    normalizer = GRN_NORMALIZER_AUTO;
-  }
-  tokenizer->nstr = grn_string_open_(ctx,
-                                     GRN_TEXT_VALUE(str), GRN_TEXT_LEN(str),
-                                     normalizer, nflags, tokenizer->encoding);
-  if (!tokenizer->nstr) {
-    GRN_FREE(tokenizer);
-    ERR(GRN_TOKENIZER_ERROR, "grn_string_open failed at grn_token_open");
-    return NULL;
-  }
-  grn_string_get_normalized(ctx, tokenizer->nstr,
+  grn_string_get_normalized(ctx, tokenizer->query->normalized_query,
                             &normalized, &normalized_length_in_bytes,
                             NULL);
   tokenizer->next = (const unsigned char *)normalized;
@@ -152,25 +163,36 @@ delimited_next(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data
     unsigned int rest_length;
     rest_length = tokenizer->end - tokenizer->next;
     tokenizer->next =
-      grn_tokenizer_tokenized_delimiter_next(ctx,
-                                             &(tokenizer->token),
-                                             tokenizer->next,
-                                             rest_length,
-                                             tokenizer->encoding);
+      (unsigned char *)grn_tokenizer_tokenized_delimiter_next(
+        ctx,
+        &(tokenizer->token),
+        (const char *)tokenizer->next,
+        rest_length,
+        tokenizer->query->encoding);
   } else {
     size_t cl;
     const unsigned char *p = tokenizer->next, *r;
     const unsigned char *e = tokenizer->end;
     grn_tokenizer_status status;
     for (r = p; r < e; r += cl) {
-      if (!(cl = grn_charlen_(ctx, (char *)r, (char *)e, tokenizer->encoding))) {
+      if (!(cl = grn_charlen_(ctx, (char *)r, (char *)e,
+                              tokenizer->query->encoding))) {
         tokenizer->next = (unsigned char *)e;
         break;
       }
-      if (r + tokenizer->delimiter_len <= e &&
-          !memcmp(r, tokenizer->delimiter, tokenizer->delimiter_len)) {
-        tokenizer->next = r + tokenizer->delimiter_len;
-        break;
+      {
+        grn_bool found_delimiter = GRN_FALSE;
+        const unsigned char *current_end = r;
+        while (current_end + tokenizer->delimiter_len <= e &&
+               !memcmp(current_end,
+                       tokenizer->delimiter, tokenizer->delimiter_len)) {
+          current_end += tokenizer->delimiter_len;
+          tokenizer->next = current_end;
+          found_delimiter = GRN_TRUE;
+        }
+        if (found_delimiter) {
+          break;
+        }
       }
     }
     if (r == e) {
@@ -178,7 +200,11 @@ delimited_next(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data
     } else {
       status = GRN_TOKENIZER_CONTINUE;
     }
-    grn_tokenizer_token_push(ctx, &(tokenizer->token), p, r - p, status);
+    grn_tokenizer_token_push(ctx,
+                             &(tokenizer->token),
+                             (const char *)p,
+                             r - p,
+                             status);
   }
 
   return NULL;
@@ -188,7 +214,10 @@ static grn_obj *
 delimited_fin(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data)
 {
   grn_delimited_tokenizer *tokenizer = user_data->ptr;
-  grn_obj_close(ctx, tokenizer->nstr);
+  if (!tokenizer) {
+    return NULL;
+  }
+  grn_tokenizer_query_close(ctx, tokenizer->query);
   grn_tokenizer_token_fin(ctx, &(tokenizer->token));
   GRN_FREE(tokenizer);
   return NULL;
@@ -198,22 +227,21 @@ static grn_obj *
 delimit_init(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data)
 {
   static const uint8_t delimiter[1] = {' '};
-  grn_obj *table = args[0];
-  return delimited_init(ctx, table, user_data, delimiter, 1);
+  return delimited_init(ctx, nargs, args, user_data, delimiter, 1);
 }
 
 static grn_obj *
 delimit_null_init(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data)
 {
   static const uint8_t delimiter[1] = {'\0'};
-  grn_obj *table = args[0];
-  return delimited_init(ctx, table, user_data, delimiter, 1);
+  return delimited_init(ctx, nargs, args, user_data, delimiter, 1);
 }
 
 /* ngram tokenizer */
 
 typedef struct {
-  grn_obj *nstr;
+  grn_tokenizer_token token;
+  grn_tokenizer_query *query;
   uint8_t uni_alpha;
   uint8_t uni_digit;
   uint8_t uni_symbol;
@@ -222,203 +250,221 @@ typedef struct {
   uint8_t overlap;
   int32_t pos;
   uint32_t skip;
-  grn_encoding encoding;
   const unsigned char *next;
   const unsigned char *end;
   const uint_least8_t *ctypes;
-  int32_t len;
+  uint32_t len;
   uint32_t tail;
-  grn_obj curr_;
-  grn_obj stat_;
 } grn_ngram_tokenizer;
 
 static grn_obj *
-ngram_init(grn_ctx *ctx, grn_obj *table, grn_user_data *user_data, uint8_t ngram_unit,
+ngram_init(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data, uint8_t ngram_unit,
            uint8_t uni_alpha, uint8_t uni_digit, uint8_t uni_symbol, uint8_t ignore_blank)
 {
-  grn_obj *str;
-  grn_obj *normalizer = NULL;
-  int nflags =
+  unsigned int normalize_flags =
     GRN_STRING_REMOVE_BLANK |
     GRN_STRING_WITH_TYPES |
     GRN_STRING_REMOVE_TOKENIZED_DELIMITER;
+  grn_tokenizer_query *query;
   const char *normalized;
   unsigned int normalized_length_in_bytes;
-  grn_ngram_tokenizer *token;
-  grn_obj_flags table_flags;
-  if (!(str = grn_ctx_pop(ctx))) {
-    ERR(GRN_INVALID_ARGUMENT, "missing argument");
+  grn_ngram_tokenizer *tokenizer;
+
+  query = grn_tokenizer_query_open(ctx, nargs, args, normalize_flags);
+  if (!query) {
     return NULL;
   }
-  if (!(token = GRN_MALLOC(sizeof(grn_ngram_tokenizer)))) { return NULL; }
-  user_data->ptr = token;
-  token->uni_alpha = uni_alpha;
-  token->uni_digit = uni_digit;
-  token->uni_symbol = uni_symbol;
-  token->ngram_unit = ngram_unit;
-  token->ignore_blank = ignore_blank;
-  token->overlap = 0;
-  token->pos = 0;
-  token->skip = 0;
-  grn_table_get_info(ctx, table, &table_flags, &token->encoding, NULL);
-  if (table_flags & GRN_OBJ_KEY_NORMALIZE) {
-    normalizer = GRN_NORMALIZER_AUTO;
-  }
-  if (!(token->nstr = grn_string_open_(ctx,
-                                       GRN_TEXT_VALUE(str), GRN_TEXT_LEN(str),
-                                       normalizer, nflags, token->encoding))) {
-    GRN_FREE(token);
-    ERR(GRN_TOKENIZER_ERROR, "grn_string_open failed at grn_token_open");
+
+  if (!(tokenizer = GRN_MALLOC(sizeof(grn_ngram_tokenizer)))) {
+    grn_tokenizer_query_close(ctx, query);
+    ERR(GRN_NO_MEMORY_AVAILABLE,
+        "[tokenizer][ngram] "
+        "memory allocation to grn_ngram_tokenizer failed");
     return NULL;
   }
-  grn_string_get_normalized(ctx, token->nstr,
+  user_data->ptr = tokenizer;
+
+  grn_tokenizer_token_init(ctx, &(tokenizer->token));
+  tokenizer->query = query;
+
+  tokenizer->uni_alpha = uni_alpha;
+  tokenizer->uni_digit = uni_digit;
+  tokenizer->uni_symbol = uni_symbol;
+  tokenizer->ngram_unit = ngram_unit;
+  tokenizer->ignore_blank = ignore_blank;
+  tokenizer->overlap = 0;
+  tokenizer->pos = 0;
+  tokenizer->skip = 0;
+
+  grn_string_get_normalized(ctx, tokenizer->query->normalized_query,
                             &normalized, &normalized_length_in_bytes,
-                            &(token->len));
-  token->next = (const unsigned char *)normalized;
-  token->end = token->next + normalized_length_in_bytes;
-  token->ctypes = grn_string_get_types(ctx, token->nstr);
-  GRN_TEXT_INIT(&token->curr_, GRN_OBJ_DO_SHALLOW_COPY);
-  GRN_UINT32_INIT(&token->stat_, 0);
+                            &(tokenizer->len));
+  tokenizer->next = (const unsigned char *)normalized;
+  tokenizer->end = tokenizer->next + normalized_length_in_bytes;
+  tokenizer->ctypes =
+    grn_string_get_types(ctx, tokenizer->query->normalized_query);
   return NULL;
 }
 
 static grn_obj *
 unigram_init(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data)
-{ return ngram_init(ctx, args[0], user_data, 1, 1, 1, 1, 0); }
+{ return ngram_init(ctx, nargs, args, user_data, 1, 1, 1, 1, 0); }
 
 static grn_obj *
 bigram_init(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data)
-{ return ngram_init(ctx, args[0], user_data, 2, 1, 1, 1, 0); }
+{ return ngram_init(ctx, nargs, args, user_data, 2, 1, 1, 1, 0); }
 
 static grn_obj *
 trigram_init(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data)
-{ return ngram_init(ctx, args[0], user_data, 3, 1, 1, 1, 0); }
+{ return ngram_init(ctx, nargs, args, user_data, 3, 1, 1, 1, 0); }
 
 static grn_obj *
 bigrams_init(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data)
-{ return ngram_init(ctx, args[0], user_data, 2, 1, 1, 0, 0); }
+{ return ngram_init(ctx, nargs, args, user_data, 2, 1, 1, 0, 0); }
 
 static grn_obj *
 bigramsa_init(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data)
-{ return ngram_init(ctx, args[0], user_data, 2, 0, 1, 0, 0); }
+{ return ngram_init(ctx, nargs, args, user_data, 2, 0, 1, 0, 0); }
 
 static grn_obj *
 bigramsad_init(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data)
-{ return ngram_init(ctx, args[0], user_data, 2, 0, 0, 0, 0); }
+{ return ngram_init(ctx, nargs, args, user_data, 2, 0, 0, 0, 0); }
 
 static grn_obj *
 bigrami_init(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data)
-{ return ngram_init(ctx, args[0], user_data, 2, 1, 1, 1, 1); }
+{ return ngram_init(ctx, nargs, args, user_data, 2, 1, 1, 1, 1); }
 
 static grn_obj *
 bigramis_init(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data)
-{ return ngram_init(ctx, args[0], user_data, 2, 1, 1, 0, 1); }
+{ return ngram_init(ctx, nargs, args, user_data, 2, 1, 1, 0, 1); }
 
 static grn_obj *
 bigramisa_init(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data)
-{ return ngram_init(ctx, args[0], user_data, 2, 0, 1, 0, 1); }
+{ return ngram_init(ctx, nargs, args, user_data, 2, 0, 1, 0, 1); }
 
 static grn_obj *
 bigramisad_init(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data)
-{ return ngram_init(ctx, args[0], user_data, 2, 0, 0, 0, 1); }
+{ return ngram_init(ctx, nargs, args, user_data, 2, 0, 0, 0, 1); }
 
 static grn_obj *
 ngram_next(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data)
 {
   size_t cl;
-  grn_ngram_tokenizer *token = user_data->ptr;
-  const unsigned char *p = token->next, *r = p, *e = token->end;
-  int32_t len = 0, pos = token->pos + token->skip, status = 0;
-  const uint_least8_t *cp = token->ctypes ? token->ctypes + pos : NULL;
-  if (cp && token->uni_alpha && GRN_STR_CTYPE(*cp) == grn_str_alpha) {
-    while ((cl = grn_charlen_(ctx, (char *)r, (char *)e, token->encoding))) {
+  grn_ngram_tokenizer *tokenizer = user_data->ptr;
+  const unsigned char *p = tokenizer->next, *r = p, *e = tokenizer->end;
+  int32_t len = 0, pos = tokenizer->pos + tokenizer->skip, status = 0;
+  const uint_least8_t *cp = tokenizer->ctypes ? tokenizer->ctypes + pos : NULL;
+  if (cp && tokenizer->uni_alpha && GRN_STR_CTYPE(*cp) == GRN_CHAR_ALPHA) {
+    while ((cl = grn_charlen_(ctx, (char *)r, (char *)e,
+                              tokenizer->query->encoding))) {
       len++;
       r += cl;
-      if (/* !token->ignore_blank && */ GRN_STR_ISBLANK(*cp)) { break; }
-      if (GRN_STR_CTYPE(*++cp) != grn_str_alpha) { break; }
+      if (/* !tokenizer->ignore_blank && */ GRN_STR_ISBLANK(*cp)) { break; }
+      if (GRN_STR_CTYPE(*++cp) != GRN_CHAR_ALPHA) { break; }
     }
-    token->next = r;
-    token->overlap = 0;
-  } else if (cp && token->uni_digit && GRN_STR_CTYPE(*cp) == grn_str_digit) {
-    while ((cl = grn_charlen_(ctx, (char *)r, (char *)e, token->encoding))) {
+    tokenizer->next = r;
+    tokenizer->overlap = 0;
+  } else if (cp &&
+             tokenizer->uni_digit &&
+             GRN_STR_CTYPE(*cp) == GRN_CHAR_DIGIT) {
+    while ((cl = grn_charlen_(ctx, (char *)r, (char *)e,
+                              tokenizer->query->encoding))) {
       len++;
       r += cl;
-      if (/* !token->ignore_blank && */ GRN_STR_ISBLANK(*cp)) { break; }
-      if (GRN_STR_CTYPE(*++cp) != grn_str_digit) { break; }
+      if (/* !tokenizer->ignore_blank && */ GRN_STR_ISBLANK(*cp)) { break; }
+      if (GRN_STR_CTYPE(*++cp) != GRN_CHAR_DIGIT) { break; }
     }
-    token->next = r;
-    token->overlap = 0;
-  } else if (cp && token->uni_symbol && GRN_STR_CTYPE(*cp) == grn_str_symbol) {
-    while ((cl = grn_charlen_(ctx, (char *)r, (char *)e, token->encoding))) {
+    tokenizer->next = r;
+    tokenizer->overlap = 0;
+  } else if (cp &&
+             tokenizer->uni_symbol &&
+             GRN_STR_CTYPE(*cp) == GRN_CHAR_SYMBOL) {
+    while ((cl = grn_charlen_(ctx, (char *)r, (char *)e,
+                              tokenizer->query->encoding))) {
       len++;
       r += cl;
-      if (!token->ignore_blank && GRN_STR_ISBLANK(*cp)) { break; }
-      if (GRN_STR_CTYPE(*++cp) != grn_str_symbol) { break; }
+      if (!tokenizer->ignore_blank && GRN_STR_ISBLANK(*cp)) { break; }
+      if (GRN_STR_CTYPE(*++cp) != GRN_CHAR_SYMBOL) { break; }
     }
-    token->next = r;
-    token->overlap = 0;
+    tokenizer->next = r;
+    tokenizer->overlap = 0;
   } else {
 #ifdef PRE_DEFINED_UNSPLIT_WORDS
     const unsigned char *key = NULL;
     // todo : grn_pat_lcp_search
     if ((tid = grn_sym_common_prefix_search(sym, p))) {
       if (!(key = _grn_sym_key(sym, tid))) {
-        token->status = grn_token_not_found;
+        tokenizer->status = GRN_TOKEN_NOT_FOUND;
         return NULL;
       }
-      len = grn_str_len(key, token->encoding, NULL);
+      len = grn_str_len(key, tokenizer->query->encoding, NULL);
     }
-    r = p + grn_charlen_(ctx, p, e, token->encoding);
+    r = p + grn_charlen_(ctx, p, e, tokenizer->query->encoding);
     if (tid && (len > 1 || r == p)) {
-      if (r != p && pos + len - 1 <= token->tail) { continue; }
+      if (r != p && pos + len - 1 <= tokenizer->tail) { continue; }
       p += strlen(key);
-      if (!*p && token->mode == grn_token_get) { token->status = grn_token_done; }
+      if (!*p && tokenizer->mode == GRN_TOKEN_GET) {
+        tokenizer->status = GRN_TOKEN_DONE;
+      }
     }
 #endif /* PRE_DEFINED_UNSPLIT_WORDS */
-    if ((cl = grn_charlen_(ctx, (char *)r, (char *)e, token->encoding))) {
+    if ((cl = grn_charlen_(ctx, (char *)r, (char *)e,
+                           tokenizer->query->encoding))) {
       len++;
       r += cl;
-      token->next = r;
-      while (len < token->ngram_unit &&
-             (cl = grn_charlen_(ctx, (char *)r, (char *)e, token->encoding))) {
+      tokenizer->next = r;
+      while (len < tokenizer->ngram_unit &&
+             (cl = grn_charlen_(ctx, (char *)r, (char *)e,
+                                tokenizer->query->encoding))) {
         if (cp) {
-          if (!token->ignore_blank && GRN_STR_ISBLANK(*cp)) { break; }
+          if (!tokenizer->ignore_blank && GRN_STR_ISBLANK(*cp)) { break; }
           cp++;
-          if ((token->uni_alpha && GRN_STR_CTYPE(*cp) == grn_str_alpha) ||
-              (token->uni_digit && GRN_STR_CTYPE(*cp) == grn_str_digit) ||
-              (token->uni_symbol && GRN_STR_CTYPE(*cp) == grn_str_symbol)) { break; }
+          if ((tokenizer->uni_alpha && GRN_STR_CTYPE(*cp) == GRN_CHAR_ALPHA) ||
+              (tokenizer->uni_digit && GRN_STR_CTYPE(*cp) == GRN_CHAR_DIGIT) ||
+              (tokenizer->uni_symbol && GRN_STR_CTYPE(*cp) == GRN_CHAR_SYMBOL)) {
+            break;
+          }
         }
         len++;
         r += cl;
       }
-      if (token->overlap) { status |= GRN_TOKEN_OVERLAP; }
-      if (len < token->ngram_unit) { status |= GRN_TOKEN_UNMATURED; }
-      token->overlap = (len > 1) ? 1 : 0;
+      if (tokenizer->overlap) {
+        status |= GRN_TOKENIZER_TOKEN_OVERLAP;
+      }
+      if (len < tokenizer->ngram_unit) {
+        status |= GRN_TOKENIZER_TOKEN_UNMATURED;
+      }
+      tokenizer->overlap = (len > 1) ? 1 : 0;
     }
   }
-  token->pos = pos;
-  token->len = len;
-  token->tail = pos + len - 1;
-  if (p == r || token->next == e) {
-    token->skip = 0;
-    status |= GRN_TOKEN_LAST;
+  tokenizer->pos = pos;
+  tokenizer->len = len;
+  tokenizer->tail = pos + len - 1;
+  if (p == r || tokenizer->next == e) {
+    tokenizer->skip = 0;
+    status |= GRN_TOKENIZER_TOKEN_LAST;
   } else {
-    token->skip = token->overlap ? 1 : len;
+    tokenizer->skip = tokenizer->overlap ? 1 : len;
   }
-  if (r == e) { status |= GRN_TOKEN_REACH_END; }
-  GRN_TEXT_SET_REF(&token->curr_, p, r - p);
-  GRN_UINT32_SET(ctx, &token->stat_, status);
-  grn_ctx_push(ctx, &token->curr_);
-  grn_ctx_push(ctx, &token->stat_);
+  if (r == e) { status |= GRN_TOKENIZER_TOKEN_REACH_END; }
+  grn_tokenizer_token_push(ctx,
+                           &(tokenizer->token),
+                           (const char *)p,
+                           r - p,
+                           status);
   return NULL;
 }
 
 static grn_obj *
 ngram_fin(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data)
 {
-  grn_ngram_tokenizer *token = user_data->ptr;
-  grn_obj_close(ctx, token->nstr);
-  GRN_FREE(token);
+  grn_ngram_tokenizer *tokenizer = user_data->ptr;
+  if (!tokenizer) {
+    return NULL;
+  }
+  grn_tokenizer_token_fin(ctx, &(tokenizer->token));
+  grn_tokenizer_query_close(ctx, tokenizer->query);
+  GRN_FREE(tokenizer);
   return NULL;
 }
 
@@ -427,15 +473,15 @@ ngram_fin(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data)
 grn_rc
 grn_token_init(void)
 {
-  static grn_proc _grn_uvector_tokenizer;
-  _grn_uvector_tokenizer.obj.db = NULL;
-  _grn_uvector_tokenizer.obj.id = GRN_ID_NIL;
-  _grn_uvector_tokenizer.obj.header.domain = GRN_ID_NIL;
-  _grn_uvector_tokenizer.obj.range = GRN_ID_NIL;
-  _grn_uvector_tokenizer.funcs[PROC_INIT] = uvector_init;
-  _grn_uvector_tokenizer.funcs[PROC_NEXT] = uvector_next;
-  _grn_uvector_tokenizer.funcs[PROC_FIN] = uvector_fin;
-  grn_uvector_tokenizer = (grn_obj *)&_grn_uvector_tokenizer;
+  static grn_proc _grn_token_uvector;
+  _grn_token_uvector.obj.db = NULL;
+  _grn_token_uvector.obj.id = GRN_ID_NIL;
+  _grn_token_uvector.obj.header.domain = GRN_ID_NIL;
+  _grn_token_uvector.obj.range = GRN_ID_NIL;
+  _grn_token_uvector.funcs[PROC_INIT] = uvector_init;
+  _grn_token_uvector.funcs[PROC_NEXT] = uvector_next;
+  _grn_token_uvector.funcs[PROC_FIN] = uvector_fin;
+  grn_token_uvector = (grn_obj *)&_grn_token_uvector;
   return GRN_SUCCESS;
 }
 
@@ -447,13 +493,15 @@ grn_token_fin(void)
 
 grn_token *
 grn_token_open(grn_ctx *ctx, grn_obj *table, const char *str, size_t str_len,
-               grn_token_mode mode)
+               grn_token_mode mode, unsigned int flags)
 {
   grn_token *token;
   grn_encoding encoding;
   grn_obj *tokenizer;
+  grn_obj *normalizer;
   grn_obj_flags table_flags;
-  if (grn_table_get_info(ctx, table, &table_flags, &encoding, &tokenizer)) {
+  if (grn_table_get_info(ctx, table, &table_flags, &encoding, &tokenizer,
+                         &normalizer)) {
     return NULL;
   }
   if (!(token = GRN_MALLOC(sizeof(grn_token)))) { return NULL; }
@@ -461,18 +509,20 @@ grn_token_open(grn_ctx *ctx, grn_obj *table, const char *str, size_t str_len,
   token->mode = mode;
   token->encoding = encoding;
   token->tokenizer = tokenizer;
-  token->orig = str;
+  token->orig = (const unsigned char *)str;
   token->orig_blen = str_len;
   token->curr = NULL;
   token->nstr = NULL;
   token->curr_size = 0;
   token->pos = -1;
-  token->status = grn_token_doing;
+  token->status = GRN_TOKEN_DOING;
   token->force_prefix = 0;
   if (tokenizer) {
-    grn_obj str_;
+    grn_obj str_, flags_;
     GRN_TEXT_INIT(&str_, GRN_OBJ_DO_SHALLOW_COPY);
     GRN_TEXT_SET_REF(&str_, str, str_len);
+    GRN_UINT32_INIT(&flags_, 0);
+    GRN_UINT32_SET(ctx, &flags_, flags);
     token->pctx.caller = NULL;
     token->pctx.user_data.ptr = NULL;
     token->pctx.proc = (grn_proc *)tokenizer;
@@ -480,14 +530,12 @@ grn_token_open(grn_ctx *ctx, grn_obj *table, const char *str, size_t str_len,
     token->pctx.currh = NULL;
     token->pctx.phase = PROC_INIT;
     grn_ctx_push(ctx, &str_);
+    grn_ctx_push(ctx, &flags_);
     ((grn_proc *)tokenizer)->funcs[PROC_INIT](ctx, 1, &table, &token->pctx.user_data);
+    grn_obj_close(ctx, &flags_);
     grn_obj_close(ctx, &str_);
   } else {
-    grn_obj *normalizer = NULL;
     int nflags = 0;
-    if (table_flags & GRN_OBJ_KEY_NORMALIZE) {
-      normalizer = GRN_NORMALIZER_AUTO;
-    }
     token->nstr = grn_string_open_(ctx, str, str_len,
                                    normalizer, nflags, token->encoding);
     if (token->nstr) {
@@ -513,33 +561,55 @@ grn_token_next(grn_ctx *ctx, grn_token *token)
   grn_id tid = GRN_ID_NIL;
   grn_obj *table = token->table;
   grn_obj *tokenizer = token->tokenizer;
-  while (token->status != grn_token_done) {
+  while (token->status != GRN_TOKEN_DONE) {
     if (tokenizer) {
       grn_obj *curr_, *stat_;
       ((grn_proc *)tokenizer)->funcs[PROC_NEXT](ctx, 1, &table, &token->pctx.user_data);
       stat_ = grn_ctx_pop(ctx);
       curr_ = grn_ctx_pop(ctx);
-      token->curr = GRN_TEXT_VALUE(curr_);
+      token->curr = (const unsigned char *)GRN_TEXT_VALUE(curr_);
       token->curr_size = GRN_TEXT_LEN(curr_);
       status = GRN_UINT32_VALUE(stat_);
-      token->status = ((status & GRN_TOKEN_LAST) ||
-                       (token->mode == grn_token_get && (status & GRN_TOKEN_REACH_END)))
-        ? grn_token_done : grn_token_doing;
+      token->status = ((status & GRN_TOKENIZER_TOKEN_LAST) ||
+                       (token->mode == GRN_TOKEN_GET &&
+                        (status & GRN_TOKENIZER_TOKEN_REACH_END)))
+        ? GRN_TOKEN_DONE : GRN_TOKEN_DOING;
       token->force_prefix = 0;
-      if (status & GRN_TOKEN_UNMATURED) {
-        if (status & GRN_TOKEN_OVERLAP) {
-          if (token->mode == grn_token_get) { token->pos++; continue; }
+      if (token->curr_size == 0) {
+        char tokenizer_name[GRN_TABLE_MAX_KEY_SIZE];
+        int tokenizer_name_length;
+        tokenizer_name_length =
+          grn_obj_name(ctx, token->tokenizer,
+                       tokenizer_name, GRN_TABLE_MAX_KEY_SIZE);
+        GRN_LOG(ctx, GRN_WARN,
+                "[token_next] ignore an empty token: <%.*s>: <%.*s>",
+                tokenizer_name_length, tokenizer_name,
+                token->orig_blen, token->orig);
+        continue;
+      }
+      if (token->curr_size > GRN_TABLE_MAX_KEY_SIZE) {
+        GRN_LOG(ctx, GRN_WARN,
+                "[token_next] ignore too long token. "
+                "Token must be less than or equal to %d: <%d>(<%.*s>)",
+                GRN_TABLE_MAX_KEY_SIZE,
+                token->curr_size,
+                token->curr_size, token->curr);
+        continue;
+      }
+      if (status & GRN_TOKENIZER_TOKEN_UNMATURED) {
+        if (status & GRN_TOKENIZER_TOKEN_OVERLAP) {
+          if (token->mode == GRN_TOKEN_GET) { token->pos++; continue; }
         } else {
-          if (status & GRN_TOKEN_LAST) { token->force_prefix = 1; }
+          if (status & GRN_TOKENIZER_TOKEN_LAST) { token->force_prefix = 1; }
         }
       }
     } else {
-      token->status = grn_token_done;
+      token->status = GRN_TOKEN_DONE;
     }
-    if (token->mode == grn_token_add) {
+    if (token->mode == GRN_TOKEN_ADD) {
       switch (table->header.type) {
       case GRN_TABLE_PAT_KEY :
-        if (grn_io_lock(ctx, ((grn_pat *)table)->io, 10000000)) {
+        if (grn_io_lock(ctx, ((grn_pat *)table)->io, grn_lock_timeout)) {
           tid = GRN_ID_NIL;
         } else {
           tid = grn_pat_add(ctx, (grn_pat *)table, token->curr, token->curr_size,
@@ -548,7 +618,7 @@ grn_token_next(grn_ctx *ctx, grn_token *token)
         }
         break;
       case GRN_TABLE_DAT_KEY :
-        if (grn_io_lock(ctx, ((grn_dat *)table)->io, 10000000)) {
+        if (grn_io_lock(ctx, ((grn_dat *)table)->io, grn_lock_timeout)) {
           tid = GRN_ID_NIL;
         } else {
           tid = grn_dat_add(ctx, (grn_dat *)table, token->curr, token->curr_size,
@@ -557,7 +627,7 @@ grn_token_next(grn_ctx *ctx, grn_token *token)
         }
         break;
       case GRN_TABLE_HASH_KEY :
-        if (grn_io_lock(ctx, ((grn_hash *)table)->io, 10000000)) {
+        if (grn_io_lock(ctx, ((grn_hash *)table)->io, grn_lock_timeout)) {
           tid = GRN_ID_NIL;
         } else {
           tid = grn_hash_add(ctx, (grn_hash *)table, token->curr, token->curr_size,
@@ -593,8 +663,8 @@ grn_token_next(grn_ctx *ctx, grn_token *token)
         break;
       }
     }
-    if (tid == GRN_ID_NIL && token->status != grn_token_done) {
-      token->status = grn_token_not_found;
+    if (tid == GRN_ID_NIL && token->status != GRN_TOKEN_DONE) {
+      token->status = GRN_TOKEN_NOT_FOUND;
     }
     token->pos++;
     break;
@@ -627,7 +697,18 @@ grn_db_init_mecab_tokenizer(grn_ctx *ctx)
   case GRN_ENC_EUC_JP :
   case GRN_ENC_UTF8 :
   case GRN_ENC_SJIS :
-    return grn_plugin_register(ctx, "tokenizers/mecab");
+    {
+      const char *mecab_plugin_name = "tokenizers/mecab";
+      char *path;
+      path = grn_plugin_find_path(ctx, mecab_plugin_name);
+      if (path) {
+        GRN_FREE(path);
+        return grn_plugin_register(ctx, mecab_plugin_name);
+      } else {
+        return GRN_NO_SUCH_FILE_OR_DIRECTORY;
+      }
+    }
+    break;
   default :
     return GRN_OPERATION_NOT_SUPPORTED;
   }
